MODULE RecAsg;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Lib, Lib2, Lib3, Str, IO, SO, Scnr;
FROM Storage IMPORT ALLOCATE;

CONST
  lenMaxRec = 32 * 1024;

  stkMax = 64;
  lenFld = 8;
  cntMax = 128;

  cSep = '.';

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. MAX( TIdx ) ];

  TFldStr = ARRAY [0..lenFld-1] OF CHAR;

  TRec = POINTER TO TRecDesc;
  TRecDesc = RECORD
    idx: TIdx;
    str: TFldStr;
    ofs: CARDINAL;
    len: CARDINAL;
    dim: CARDINAL;  (* 0: no array, >0: array dimension *)
    fst: TRec;      (* first field of this record, NIL = empty or data *)
    nxt: TRec;      (* next field of parent record *)
  END;

VAR
  arr: ARRAY TVldIdx OF TRec;
  cnt: TCnt;
  root: TRec;
  buf: POINTER TO ARRAY [0..lenMaxRec-1] OF BYTE;
  len: CARDINAL;


PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Lib3.FatalError( s );
END Error;


PROCEDURE Alp( c: CHAR ): BOOLEAN;
BEGIN
  RETURN ( 'A' <= c ) & ( c <= 'Z' ) OR
         ( 'a' <= c ) & ( c <= 'z' ) OR
         ( c = '_' );
END Alp;

PROCEDURE AlpNum( c: CHAR ): BOOLEAN;
BEGIN
  RETURN Alp( c ) OR ( '0' <= c ) & ( c <= '9' );
END AlpNum;


PROCEDURE New( s-: ARRAY OF CHAR ): TRec;
VAR r: TRec;
BEGIN
  IF cnt >= cntMax THEN Error( 'New: Too many records' ) END;
  NEW( r );
  Lib2.Fill( r^, 0 );
  INC( cnt );
  r^.idx := cnt;
  Str.Copy( r^.str, s );
  arr[ cnt ] := r;
  RETURN r;
END New;

PROCEDURE Idx( r: TRec ): TIdx;
BEGIN
  IF r = NIL THEN
    RETURN 0;
  ELSE
    RETURN r^.idx;
  END;
END Idx;

PROCEDURE Rec( i: TIdx ): TRec;
BEGIN
  IF ( i = 0 ) OR ( i > cnt ) THEN RETURN NIL END;
  RETURN arr[ i ];
END Rec;

PROCEDURE ShowDef();

  PROCEDURE Hdg();
  BEGIN
    IO.WrStr( 'idx str         ofs len dim fst nxt' );
    IO.WrLn();
  END Hdg;

  PROCEDURE Rec( r: TRec );
  VAR n: CARDINAL;
  BEGIN
    IO.WrCard( r^.idx, 3 );
    IO.WrStr( " '" );
    IO.WrStr( r^.str );
    IO.WrStr( "' " );
    n := lenFld - Str.Length( r^.str );
    WHILE n > 0 DO IO.WrChar( ' ' ); DEC( n ) END;
    IO.WrCard( r^.ofs, 4 );
    IO.WrCard( r^.len, 4 );
    IO.WrCard( r^.dim, 4 );
    IO.WrCard( Idx( r^.fst ), 4 );
    IO.WrCard( Idx( r^.nxt ), 4 );
    IO.WrLn();
  END Rec;

  PROCEDURE List();
  VAR i: TIdx;
  BEGIN
    Hdg();
    FOR i := 1 TO cnt DO
      Rec( arr[i] );
    END;
  END List;

  PROCEDURE Inspect( r: TRec );
  BEGIN
    IF r = NIL THEN RETURN END;
    Rec( r );
    r := r^.fst;
    WHILE r # NIL DO
      Inspect( r );
      r := r^.nxt;
    END;
  END Inspect;

BEGIN
  Hdg();
  Inspect( root );
  (*
  List();
  *)
END ShowDef;


PROCEDURE Build();
VAR
  rec: TRec;  (* active record *)
  pos: TRec;  (* position in active record *)
  ofs: CARDINAL;
  stk: RECORD
    n: CARDINAL;
    a: ARRAY [0..stkMax-1] OF TRec;
  END;

  PROCEDURE ShowOp( s-, t-: ARRAY OF CHAR );
  CONST ena = FALSE;
  VAR n: CARDINAL;
  BEGIN
    IF ~ ena THEN RETURN END;
    IO.WrStr( s );
    IO.WrStr( " '" );
    IO.WrStr( t );
    IO.WrStr( "' " );
    n := lenFld - Str.Length( t );
    WHILE n > 0 DO IO.WrChar( ' ' ); DEC( n ) END;
    IO.WrCard( Idx( rec ), 3 );
    IO.WrCard( Idx( pos ), 3 );
    IO.WrLn();
  END ShowOp;

  PROCEDURE ShowStk();
  VAR i: CARDINAL;
  BEGIN
    ShowDef();
    IO.WrStr( 'rec ' );
    IO.WrCard( Idx( rec ), 1 );
    IO.WrStr( '  pos ' );
    IO.WrCard( Idx( pos ), 1 );
    IO.WrStr( '  ofs ' );
    IO.WrCard( ofs, 1 );
    IO.WrLn();
    IO.WrStr( 'stk.n ' );
    IO.WrCard( stk.n, 1 );
    IF stk.n > 0 THEN
      IO.WrStr( '  stk.a' );
      FOR i := 0 TO stk.n - 1 DO
        IO.WrChar( ' ' );
        IO.WrCard( Idx( stk.a[i] ), 1 );
      END;
    END;
    IO.WrLn();
  END ShowStk;

  PROCEDURE WrFull( VAR b: SO.TBuffer; s-: ARRAY OF CHAR );
  VAR i: CARDINAL;
  BEGIN
    i := 1;
    WHILE i < stk.n DO
      SO.WrStr( b, stk.a[i]^.str );
      SO.WrChar( b, cSep );
      INC( i );
    END;
    IF rec # root THEN
      SO.WrStr( b, rec^.str );
      SO.WrChar( b, cSep );
    END;
    SO.WrStr( b, s );
  END WrFull;

  PROCEDURE Err( x-, y-: ARRAY OF CHAR );
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, 'Build.' );
    SO.WrStr( b, x );
    IF y[0] # 0C THEN
      SO.WrStr( b, ": '" );
      SO.WrStr( b, y );
      SO.WrStr( b, "'" );
    END;
    Error( s );
  END Err;

  PROCEDURE ErrFull( x-, y-: ARRAY OF CHAR );
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, 'Build.' );
    SO.WrStr( b, x );
    IF y[0] # 0C THEN
      SO.WrStr( b, ": '" );
      WrFull( b, y );
      SO.WrStr( b, "'" );
    END;
    Error( s );
  END ErrFull;

  PROCEDURE Chk( s-: ARRAY OF CHAR );

    PROCEDURE ChrOk( c: CHAR; i: CARDINAL ): BOOLEAN;
    BEGIN
      IF i = 0 THEN
        RETURN Alp( c );
      ELSE
        RETURN AlpNum( c );
      END;
    END ChrOk;

    PROCEDURE Name( s-: ARRAY OF CHAR );
    VAR i: CARDINAL;
    BEGIN
      i := 0;
      WHILE ( i < SIZE( s ) ) & ( s[i] # 0C ) DO
        IF ~ ChrOk( s[i], i ) THEN Err( 'Chk: Invalid field name', s ) END;
        INC( i );
      END;
    END Name;

    PROCEDURE Fields( s-: ARRAY OF CHAR );
    VAR r: TRec;
    BEGIN
      r := rec^.fst;
      WHILE r # NIL DO
        IF Str.Compare( r^.str, s ) = 0 THEN ErrFull( 'Chk: Field already exists', s ) END;
        r := r^.nxt;
      END;
    END Fields;

  BEGIN
    Name( s );
    Fields( s );
  END Chk;

  PROCEDURE Enter( s-: ARRAY OF CHAR );
  BEGIN
    IF stk.n >= stkMax THEN ErrFull( 'Enter: Stack exhausted', s ) END;
    stk.a[ stk.n ] := rec;
    INC( stk.n );
    rec := pos;
  END Enter;

  PROCEDURE Leave( s-: ARRAY OF CHAR );
  BEGIN
    IF stk.n = 0 THEN Err( 'Leave: Stack empty', s ) END;
    pos := rec;
    DEC( stk.n );
    rec := stk.a[ stk.n ];
  END Leave;

  PROCEDURE Link( r: TRec );
  BEGIN
    IF pos = rec THEN
      pos^.fst := r;
    ELSE
      pos^.nxt := r;
    END;
    pos := r;
  END Link;

  PROCEDURE Beg( s-: ARRAY OF CHAR );
  VAR r: TRec;
  BEGIN
    ShowOp( 'BegBeg', s );
    Chk( s );
    r := New( s );
    r^.ofs := ofs;
    Link( r );
    Enter( s );
    ShowOp( 'BegEnd', s );
  END Beg;

  PROCEDURE Add( s-: ARRAY OF CHAR; d: CARDINAL; l: CARDINAL );
  VAR r: TRec;
  BEGIN
    ShowOp( 'AddBeg', s );
    Chk( s );
    IF d > 0 THEN l := d * l END;
    r := New( s );
    r^.ofs := ofs;
    r^.len := l;
    r^.dim := d;
    Link( r );
    INC( rec^.len, l );
    INC( ofs, l );
    ShowOp( 'AddEnd', s );
  END Add;

  PROCEDURE End( s-: ARRAY OF CHAR; d: CARDINAL );
  VAR l: CARDINAL;
  BEGIN
    ShowOp( 'EndBeg', s );
    IF rec = root THEN Err( 'End: No active record', s ) END;
    Leave( s );
    l := pos^.len;
    IF d > 0 THEN l := d * l END;
    pos^.len := l;
    pos^.dim := d;
    INC( rec^.len, l );
    IF Str.Compare( pos^.str, s ) # 0 THEN ErrFull( 'End: Name mismatch', s ) END;
    ShowOp( 'EndEnd', s );
  END End;

BEGIN
  rec := root;
  pos := rec;
  ofs := 0;
  Lib2.Fill( stk, 0 );

  Beg( 'tel' );
    Add( 'a', 0, 1 );
    Beg( 'b' );
      Add( 'c', 0, 2 );
    End( 'b', 0 );
    Add( 'd', 0, 4 );
    Beg( 'e' );
      Add( 'f', 0, 2 );
    End( 'e', 12 );
    Add( 'e0', 0, 1 );
  End( 'tel', 0 );
  Add( 'bla', 2, 2 );

  IF rec # root THEN Error( 'Build: Incomplete record' ) END;

  len := root^.len;
  IF len > lenMaxRec THEN Error( 'Build: Root record too large' ) END;
  ALLOCATE( buf, len );
  Lib.Fill( buf, len, 0C );
END Build;


PROCEDURE Assign();

  PROCEDURE Err( t-: ARRAY OF CHAR );
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, 'Assign: @(' );
    SO.WrLngCard( b, Scnr.pos.l, 1 );
    SO.WrChar( b, ',' );
    SO.WrLngCard( b, Scnr.pos.c, 1 );
    SO.WrStr( b, '): ' );
    SO.WrStr( b, t );
    Error( s );
  END Err;

  PROCEDURE Find( b: TRec; s-: ARRAY OF CHAR ): TRec;
  VAR r: TRec;
  BEGIN
    r := b^.fst;
    WHILE r # NIL DO
      IF Str.Compare( r^.str, s ) = 0 THEN RETURN r END;
      r := r^.nxt;
    END;
    RETURN NIL;
  END Find;

  PROCEDURE Assignment();
  VAR
    r: TRec;
    o: CARDINAL;

    PROCEDURE Ident();
    BEGIN
      IF Scnr.sym # Scnr.symId THEN Err( 'Identifier expected' ) END;
      r := Find( r, Scnr.lit );
      IF r = NIL THEN Err( 'Undeclared identifier' ) END;
      Scnr.Read();
      IF Scnr.sym = Scnr.symLBrk THEN
        IF r^.dim = 0 THEN Err( 'Not an array' ) END;
        Scnr.Read();
        IF Scnr.sym # Scnr.symNum THEN Err( 'Number expected' ) END;
        IF Scnr.num >= r^.dim THEN Err( 'Index out of range' ) END;
        INC( o, Scnr.num * ( r^.len DIV r^.dim ) );
        Scnr.Read();
        IF Scnr.sym # Scnr.symRBrk THEN Err( 'Expected "]"' ) END;
        Scnr.Read();
      END;
    END Ident;

    PROCEDURE LenNum( x: Scnr.TNumber ): CARDINAL;
    VAR l: CARDINAL;
    BEGIN
      l := 0;
      REPEAT x := x DIV 100H; INC( l ) UNTIL x = 0;
      RETURN l;
    END LenNum;

  VAR
    l: CARDINAL;
  BEGIN
    r := root;
    o := 0;
    Ident();
    WHILE Scnr.sym = Scnr.symDot DO
      Scnr.Read();
      Ident();
    END;
    IF Scnr.sym # Scnr.symBec THEN Err( 'Expected ":="' ) END;
    Scnr.Read();
    CASE Scnr.sym OF
    | Scnr.symNum:
        l := LenNum( Scnr.num );
        IF l > r^.len THEN Err( 'Number too large' ) END;
        Lib.Move( ADR(Scnr.num), ADR(buf^[r^.ofs+o]), l );
    | Scnr.symLit:
        Err( 'String assignment not yet implemented' );
    ELSE
        Err( 'Number or literal expected' )
    END;
    Scnr.Read();
    IF Scnr.sym # Scnr.symSemi THEN Err( 'Semicolon expected' ) END;
    Scnr.Read();
  END Assignment;

  PROCEDURE Show();
  BEGIN
  END Show;

BEGIN
  Scnr.Read();
  WHILE Scnr.sym = Scnr.symId DO Assignment() END;
  IF Scnr.sym # Scnr.symEOS THEN Err( 'Identifier or EOF expected' ) END;
  Show();
END Assign;


PROCEDURE Show();
VAR
  str: ARRAY [0..255] OF CHAR;
  pos: CARDINAL;

  PROCEDURE PutC( c: CHAR );
  BEGIN
    IF pos >= SIZE( str ) THEN Error( 'Show: String exhausted' ) END;
    str[pos] := c;
    INC( pos );
  END PutC;

  PROCEDURE PutS( s-: ARRAY OF CHAR );
  VAR i: CARDINAL;
  BEGIN
    i := 0;
    WHILE ( i < SIZE(s) ) & ( s[i] # 0C ) DO PutC( s[i] ); INC( i ) END;
  END PutS;

  PROCEDURE Put( s-: ARRAY OF CHAR );
  BEGIN
    IF pos > 0 THEN PutC( '.' ) END;
    PutS( s );
  END Put;

  PROCEDURE PutIdx( t-: ARRAY OF CHAR; i: CARDINAL );
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..15] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrLngCard( b, i, 1 );
    Put( t );
    PutC( '[' );
    PutS( s );
    PutC( ']' );
  END PutIdx;

  PROCEDURE Print( c: CHAR );
  VAR i: CARDINAL;
  BEGIN
    IO.WrChar( c );
    IO.WrChar( ' ' );
    i := 0;
    WHILE i < pos DO IO.WrChar( str[i] ); INC( i ) END;
    IO.WrLn();
  END Print;

  PROCEDURE Inspect( r: TRec );

    PROCEDURE Do( r: TRec );
    VAR
      old: CARDINAL;
    BEGIN
      old := pos;
      Put( r^.str );
      r := r^.fst;
      IF r = NIL THEN Print( 'S' ); pos := old; RETURN END;
      REPEAT
        Inspect( r );
        r := r^.nxt;
      UNTIL r = NIL;
      pos := old;
    END Do;

    PROCEDURE DoIdx( r: TRec; i: CARDINAL );
    VAR
      old: CARDINAL;
    BEGIN
      old := pos;
      PutIdx( r^.str, i );
      r := r^.fst;
      IF r = NIL THEN Print( 'A' ); pos := old; RETURN END;
      REPEAT
        Inspect( r );
        r := r^.nxt;
      UNTIL r = NIL;
      pos := old;
    END DoIdx;

  VAR
    i: CARDINAL;
  BEGIN
    IF r = NIL THEN RETURN END;
    IF r^.dim = 0 THEN
      Do( r );
    ELSE
      FOR i := 0 TO r^.dim - 1 DO DoIdx( r, i ) END;
    END;
  END Inspect;

BEGIN
  pos := 0;
  Inspect( root );
END Show;


BEGIN
  Lib2.Fill( arr, 0 );
  cnt := 0;
  root := New( '/' );
  buf := NIL;
  len := 0;

  Build();
  ShowDef();

  Assign();
  Show();
END RecAsg.
