(* record assignment scanner *)
IMPLEMENTATION MODULE Scnr;

IMPORT Lib2, Lib3, FIO, FIOStd;
IMPORT SO;

CONST
  cLF  = 12C;
  cCR  = 15C;
  bufLen = 4 * 1024;

  noFile ::= FIOStd.inv;

TYPE
  TBufDesc = RECORD
    n: CARDINAL;
    i: CARDINAL;
    d: ARRAY [ 1 .. bufLen ] OF CHAR;
  END;

VAR
  src: FIO.File;
  buf: TBufDesc;
  chr: CHAR;
  eof: BOOLEAN;
  chp: TPosDesc;


PROCEDURE Error( t-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'Scnr: ' );
  IF src # noFile THEN
    SO.WrStr( b, ' @(' );
    SO.WrLngCard( b, chp.l, 1 );
    SO.WrChar( b, ',' );
    SO.WrLngCard( b, chp.c, 1 );
    SO.WrStr( b, '): ' );
  END;
  SO.WrStr( b, t );
  Lib3.FatalError( s );
END Error;


PROCEDURE Get();
BEGIN
  IF eof THEN RETURN END;
  IF chr = cLF THEN
    INC( chp.l ); chp.c := 1;
  ELSE
    INC( chp.c );
  END;
  INC( buf.i );
  IF buf.i > buf.n THEN
    buf.n := FIO.RdBin( src, buf.d, SIZE(buf.d) );
    buf.i := 1;
    IF buf.n <= 0 THEN chr := 0C; eof := TRUE; RETURN END;
  END;
  chr := buf.d[ buf.i ];
  (*
  IO.WrStr( 'got ' );
  IO.WrHex( ORD(chr), 2 );
  IO.WrStr( ' @(' );
  IO.WrLngCard( chp.l, 1 );
  IO.WrChar( ',' );
  IO.WrLngCard( chp.c, 1 );
  IO.WrChar( ')' );
  IO.WrLn();
  *)
END Get;


PROCEDURE Alpha(): BOOLEAN;
BEGIN
  RETURN ( '0' <=     chr  ) & (     chr  <= '9' ) OR
         ( 'A' <= CAP(chr) ) & ( CAP(chr) <= 'Z' );
END Alpha;

PROCEDURE Read();

  PROCEDURE Put();
  BEGIN
    IF len >= SIZE( lit ) THEN Error( 'Identifier or literal too long' ) END;
    lit[ len ] := chr; INC( len );
  END Put;

  PROCEDURE Term();
  BEGIN
    IF len >= SIZE( lit ) THEN RETURN END;
    lit[ len ] := 0C;
  END Term;

  PROCEDURE Ident();
  BEGIN
    sym := symId;
    len := 0;
    REPEAT Put(); Get() UNTIL ~ Alpha();
    Term();
  END Ident;

  PROCEDURE Literal();
  VAR del: CHAR;
  BEGIN
    sym := symLit;
    len := 0;
    del := chr;
    Get();
    WHILE chr # del DO
      IF eof THEN chp := pos; Error( 'Unterminated literal (start pos)' ) END;
      IF ( chr < ' ' ) OR ( chr > '~' ) THEN Error( 'Invalid character in literal' ) END;
      Put(); Get();
    END;
    Term();
    Get();
  END Literal;

  PROCEDURE Digit( c: CHAR ): TNumber;
  BEGIN
    c := CAP( c );
    CASE c OF
    | '0'..'9': RETURN ORD( c ) - ORD( '0' );
    | 'A'..'Z': RETURN ORD( c ) - ORD( 'A' ) + 10;
    ELSE
      Error( 'Invalid digit' );
    END;
  END Digit;

  PROCEDURE Number( b: CARDINAL );
  VAR d: TNumber;
  BEGIN
    sym := symNum;
    num := 0;
    REPEAT
      d := Digit( chr );
      IF d >= b THEN Error( 'Invalid digit' ) END;
      IF num <= ( MAX(TNumber) - d ) DIV b THEN
        num := b * num + d;
      ELSE
        Error( 'Number too large' );
      END;
      Get();
    UNTIL ~ Alpha();
  END Number;

  PROCEDURE Comment();
  VAR lvl: CARDINAL;
  BEGIN
    Get(); lvl := 1;
    WHILE lvl > 0 DO
      IF eof THEN chp := pos; Error( 'Unterminated comment (start pos)' ) END;
      CASE chr OF
      | '(' : Get(); IF chr = '*' THEN INC( lvl ); Get() END;
      | '*' : Get(); IF chr = ')' THEN DEC( lvl ); Get() END;
      ELSE
        Get();
      END;
    END;
  END Comment;

BEGIN
  (* ignore control characters except eof *)
  LOOP
    IF eof THEN sym := symEOS; RETURN END;
    IF ( chr > ' ' ) & ( chr <= '~' ) THEN EXIT END;
    Get();
  END;
  pos := chp;
  CASE chr OF
  | '"',"'" : Literal();
  | '#'     : sym := symNeq; Get();
  | '$'     : Get(); Number( 16 );
  | '%'     : Get(); Number( 2 );
  | '&'     : sym := symAnd; Get();
  | '('     : Get();
              IF chr = '*' THEN Comment(); Read() ELSE sym := symLPar END;
  | ')'     : sym := symRPar; Get();
  | '*'     : sym := symMul; Get();
  | '+'     : sym := symMul; Get();
  | ','     : sym := symComma; Get();
  | '-'     : sym := symMinus; Get();
  | '.'     : sym := symDot; Get();
  | '/'     : sym := symDiv; Get();
  | '0'..'9': Number( 10 );
  | ':'     : Get();
              IF chr = '=' THEN sym := symBec; Get() ELSE sym := symCol END;
  | ';'     : sym := symSemi; Get();
  | '<'     : Get();
              CASE chr OF
              | '=': sym := symLeq; Get();
              | '>': sym := symNeq; Get();
              ELSE
                sym := symLss;
              END;
  | '='     : sym := symEql; Get();
  | '>'     : Get();
              IF chr = '=' THEN sym := symGeq; Get() ELSE sym := symGtr END;
  | 'A'..'Z': Ident();
  | '['     : sym := symLBrk; Get();
  | ']'     : sym := symRBrk; Get();
  | 'a'..'z': Ident();
  | '{'     : sym := symLBrc; Get();
  | '|'     : sym := symBar; Get();
  | '}'     : sym := symRBrc; Get();
  | '~'     : sym := symNot; Get();
  ELSE
    Error( 'Invalid character' );
  END;
END Read;

PROCEDURE Open( n-: ARRAY OF CHAR );
BEGIN
  sym := symNull;
  Lib2.Fill( pos, 0 );

  IF ~ FIO.Exists( n ) THEN Error( 'Open: Input file does not exist' ) END;
  src := FIO.OpenRead( n );
  IF src = noFile THEN Error( 'Open: Cannot open input file' ) END;

  Lib2.Fill( buf, 0 );
  chr := cLF;
  eof := FALSE;
  Lib2.Fill( chp, 0 );

  Get();
END Open;

PROCEDURE Close();
BEGIN
  IF src = noFile THEN RETURN END;
  FIO.Close( src );
  src := noFile;
END Close;

BEGIN
  sym := symNull;
  Lib2.Fill( pos, 0 );

  src := noFile;
  Lib2.Fill( buf, 0 );
  chr := 0C;
  eof := FALSE;
  Lib2.Fill( chp, 0 );

  Open( 'input.txt' );
END Scnr.
